class Error(Exception):
    """Generic error exceptions."""


class UnsupportedPlatformError(Error):
    """Raised when the platform isn't supported."""


class ParserError(Error):
    """Parser-related exceptions."""


class SteamworksError(Error):
    """Steamworks API-related exceptions."""



from dataclasses import dataclass


@dataclass(frozen=True)
class Rewards:
    bounty: int
    xp: int
    hunt_dollars: int
    bloodbonds: int
    hunter_xp: int
    hunter_levels: int
    upgrade_points: int
    bloodline_xp: int
    event_points: int

    def __bool__(self) -> bool:
        """
        Override the bool method to check if there were any rewards.
        :return: True if the match had rewards otherwise False
        """
        return any(variable_value for variable_value in vars(self).values())



from colorama import Fore, Style

from .constants import MMR_RANGES, STAR_SYMBOL


def format_mmr(mmr: int) -> str:
    """
    Formats an MMR score to a nice string.
    :param mmr: an MMR score
    :return: a color-formatted string
    """
    stars: int = len(MMR_RANGES) - len([mmr_range for mmr_range in MMR_RANGES if mmr < mmr_range])
    return f"{Fore.BLUE}{mmr}{Style.RESET_ALL} {Fore.LIGHTYELLOW_EX}{STAR_SYMBOL * stars}{Style.RESET_ALL}"



import builtins
from typing import TypeAlias, TypeVar
from xml.etree.ElementTree import Element as XmlElement

from ...exceptions import ParserError

ElementValueType: TypeAlias = str | int | bool
_T = TypeVar("_T", bound=ElementValueType)


def append_element(parent_element: XmlElement, name: str, value: ElementValueType) -> None:
    """
    Appends a new element to the parent element.
    :param parent_element: the element to append to
    :param name: the name of the attribute
    :param value: the value of the attribute
    """
    # Transform the value to a string
    match value:
        case bool():
            value = str(value).lower()
        case _:
            value = str(value)

    # Create and append the element
    new_element: XmlElement = XmlElement("Attr", attrib={"name": name, "value": value})
    parent_element.append(new_element)


# https://github.com/python/mypy/issues/3737
def get_element_value(element: XmlElement, name: str, result_type: type[_T] = str) -> _T:  # type: ignore[assignment]
    """
    Resolves an element's "value" attribute based from its name (and suffix).
    :param element: an XmlElement instance
    :param name: the element's name
    :param result_type: the type to cast the value to
    :return: the value of an element
    :raises ParserError: if the xpath isn't found,
                         if the element is missing a "value" attribute, or
                         if the result type isn't a supported type
    """
    xpath: str = f"Attr[@name={name!r}]"
    resolved_element: XmlElement | None = element.find(path=xpath)
    if resolved_element is not None:
        value: str | None = resolved_element.attrib.get("value", None)
        if value is None:
            raise ParserError(f"""Missing "value" attribute in element {xpath!r}.""")

        match result_type:
            case builtins.str | builtins.int:
                try:
                    return result_type(value)  # type: ignore[return-value]
                except ValueError:
                    raise ParserError(f"Couldn't cast the value {value!r} to an {result_type!r} type.")
            case builtins.bool:
                return value == "true"  # type: ignore[return-value]
            case _:
                raise ParserError("Type conversion not supported.")
    raise ParserError(f"No such element {xpath!r} in the current element tree.")



# Entry categories for rewards
_BOSS_NAMES: tuple[str, ...] = ("spider", "butcher", "assassin", "scrapbeak")

BOUNTY_CATEGORIES: tuple[str, ...] = (
    "accolade_bad_luck",
    "accolade_clues_found",
    *(f"accolade_found_{boss_name}" for boss_name in _BOSS_NAMES),
    *(f"accolade_killed_{boss_name}" for boss_name in _BOSS_NAMES),
    *(f"accolade_killed_{boss_name}_solo" for boss_name in _BOSS_NAMES),
    *(f"accolade_contract_{boss_name}" for boss_name in _BOSS_NAMES),
    *(f"accolade_clean_sweep_{boss_name}" for boss_name in _BOSS_NAMES),
    "accolade_trophy_extraction_bonus",
    "accolade_quickplay_found_final_clue",
    "accolade_quickplay_wellsprings_found",
    "accolade_quickplay_extracted_artifact")

XP_CATEGORIES: tuple[str, ...] = (
    "accolade_reviver",
    *(f"accolade_banished_{boss_name}" for boss_name in _BOSS_NAMES),
    "accolade_partner_killed",
    "accolade_players_killed",
    "accolade_monsters_killed")

HUNT_DOLLARS_CATEGORY: str = "accolade_found_gold"
BLOODBONDS_CATEGORY: str = "accolade_found_gems"
HUNTER_XP_DESCRIPTOR_NAME: str = "loot hunter xp"
HUNTER_XP_REWARD_TYPE: int = 10
HUNTER_LEVELS_CATEGORY: str = "accolade_hunter_points"
UPGRADE_POINTS_DESCRIPTOR_NAME: str = "loot upgrade points"
BLOODLINE_DESCRIPTOR_NAME: str = "loot bloodline xp"
ASSISTS_CATEGORY: str = "accolade_players_killed_assist"



from __future__ import annotations

from enum import IntEnum, auto as _enum_auto


class ExitCode(IntEnum):
    # No errors occurred
    SUCCESS: ExitCode = 0  # type: ignore[assignment]
    # Filesystem error
    FILESYSTEM_ERROR: ExitCode = _enum_auto()  # type: ignore[assignment]
    # Steamworks API related error
    STEAMWORKS_ERROR: ExitCode = _enum_auto()  # type: ignore[assignment]
    # Unsupported platform
    UNSUPPORTED_PLATFORM: ExitCode = _enum_auto()  # type: ignore[assignment]



from __future__ import annotations

from dataclasses import dataclass

from .xml.serializable import MappingGenerator, Serializable, XmlElement


@dataclass(frozen=True)
class Accolade(Serializable):
    bloodline_xp: int
    bounty: int
    category: str
    event_points: int
    bloodbonds: int
    generated_bloodbonds: int
    hunt_dollars: int
    hits: int
    hunter_points: int
    hunter_xp: int
    weighting: int
    xp: int

    # Serialization
    @staticmethod
    def _generate_prefix(accolade_id: int) -> str:
        """
        Generate a prefix for the class.
        :param accolade_id: the id of the accolade
        :return: a MissionAccoladeEntry prefix
        """
        return f"MissionAccoladeEntry_{accolade_id}"

    @staticmethod
    def _data_mappings() -> MappingGenerator:
        """
        Yield each variable name and its corresponding attribute suffix.
        :return: a generator which yields the name, value
        """
        # Yield the values
        yield "bloodline_xp", "bloodlineXp"
        yield "bounty", "bounty"
        yield "category", "category"
        yield "event_points", "eventPoints"
        yield "bloodbonds", "gems"
        yield "generated_bloodbonds", "generatedGems"
        yield "hunt_dollars", "gold"
        yield "hits", "hits"
        yield "hunter_points", "hunterPoints"
        yield "hunter_xp", "hunterXp"
        yield "weighting", "weighting"
        yield "xp", "xp"

    # noinspection PyMethodOverriding
    def serialize(self, root: XmlElement, *, accolade_id: int) -> None:  # type: ignore[override]
        """
        Serialize an Accolade instance.
        :param root: the root element to serialize into
        :param accolade_id: the id of the accolade
        """
        # Generate the prefix
        prefix: str = Accolade._generate_prefix(accolade_id)

        # Serialize the data
        super().serialize(root, prefix)

    # noinspection PyMethodOverriding
    @classmethod
    def deserialize(cls, root: XmlElement, accolade_id: int) -> Accolade:  # type: ignore[override]
        """
        Deserialize a series of elements into an Accolade instance.
        :param root: the root element to serialize into
        :param accolade_id: the id of the accolade
        :return: a serialized Accolade instance
        """
        # Generate the prefix
        prefix: str = Accolade._generate_prefix(accolade_id)

        # Deserialize the data
        return super(cls, cls).deserialize(root, prefix)



from contextlib import closing

from .client import Client as DatabaseClient, Cursor


def data_hash_exists(database: DatabaseClient, match_hash: str) -> bool:
    """
    Checks if a match hash already exists in the database
    :param database: a DatabaseClient instance
    :param match_hash: the hash to check for
    :return: True if a hash is already in the database, otherwise False
    """
    cursor: Cursor
    with closing(database.cursor()) as cursor:
        query: str = "SELECT EXISTS(SELECT 1 FROM data_hashes WHERE hash = ?)"
        return cursor.execute(query, (match_hash,)).fetchone()[0] >= 1


def insert_match_hash(database: DatabaseClient, match_hash: str, file_path: str) -> None:
    """
    Saves a match hash to the database.
    :param database: a DatabaseClient instance
    :param match_hash: the hash to save
    :param file_path: the path to the file
    """
    cursor: Cursor
    with closing(database.cursor()) as cursor:
        query: str = "INSERT INTO data_hashes (hash, path) VALUES (?, ?)"
        cursor.execute(query, (match_hash, file_path))
    database.save()


def update_player_data(database: DatabaseClient, profile_id: int, name: str, mmr: int,
                       kills: int, deaths: int, is_quickplay: bool) -> None:
    """
    Inserts and updates a player's data in the database.
    :param database: a DatabaseClient instance
    :param profile_id: the profile id of the player
    :param mmr: the current NNR of the player
    :param name: the name of the player
    :param kills: the amount of times the player was killed by us
    :param deaths: the amount of times we died to the player
    :param is_quickplay: True if the match was a quickplay match
    """
    # Construct the queries to execute (the strings are duplicated for easier code refactoring/highlighting)
    insert_query: str = "INSERT OR IGNORE INTO player_log_bountyhunt (profile_id, name) VALUES (?, ?)"
    update_query: str = "UPDATE player_log_bountyhunt SET name = ?, mmr = ?, " \
                        "kills = kills + ?, deaths = deaths + ?, " \
                        "encounters = encounters + 1 WHERE profile_id = ?"
    if is_quickplay:
        insert_query = "INSERT OR IGNORE INTO player_log_quickplay (profile_id, name) VALUES (?, ?)"
        update_query = "UPDATE player_log_quickplay SET name = ?, mmr = ?, " \
                       "kills = kills + ?, deaths = deaths + ?, " \
                       "encounters = encounters + 1 WHERE profile_id = ?"

    cursor: Cursor
    with closing(database.cursor()) as cursor:
        # Insert the data if it doesn't exist
        cursor.execute(insert_query, (profile_id, name))

        # Update all relevant values
        cursor.execute(update_query, (name, mmr, kills, deaths, profile_id))
    database.save()



from abc import ABC, abstractmethod
from dataclasses import is_dataclass
from typing import Generator, TypeAlias, TypeVar, get_type_hints

from .elements import ElementValueType, XmlElement, append_element, get_element_value

_T = TypeVar("_T", bound="Serializable")
MappingGenerator: TypeAlias = Generator[tuple[str, str], None, None]


class Serializable(ABC):
    @staticmethod
    @abstractmethod
    def _data_mappings() -> MappingGenerator:
        """
        Yield each variable name and its corresponding attribute suffix.
        :return: a generator which yields the name, value
        """
        raise NotImplementedError("Unimplemented _data_mappings method.")  # pragma: no cover

    @abstractmethod
    def serialize(self, root: XmlElement, name_prefix: str) -> None:
        """
        Serialize the current class into a series of XmlElement instances.
        :param root: the root element to serialize into
        :param name_prefix: the name prefix to use when appending elements
        """
        assert is_dataclass(self.__class__), "The class should be a dataclass."

        # Append each element
        for variable_name, name_suffix in self._data_mappings():
            append_element(root, name=f"{name_prefix}_{name_suffix}", value=self.__getattribute__(variable_name))

    @classmethod
    @abstractmethod
    def deserialize(cls: type[_T], root: XmlElement, name_prefix: str) -> _T:
        """
        Deserialize a series of elements into the class instance.
        :param root: the root element to deserialize from
        :param name_prefix: the name prefix to use when resolving elements
        """
        assert is_dataclass(cls), "The class should be a dataclass."

        # Generate the data by fetching each element value
        type_hints: dict[str, type[ElementValueType]] = get_type_hints(cls)
        data: dict[str, ElementValueType] = dict(
            (variable_name, get_element_value(  # type: ignore[type-var, misc]
                root, name=f"{name_prefix}_{name_suffix}", result_type=type_hints[variable_name]))
            for variable_name, name_suffix in cls._data_mappings())

        # Construct and return the class
        return cls(**data)



from __future__ import annotations

import ctypes
import os
import sys
from ctypes import Array, CDLL, cdll
from zipfile import ZipFile

from .type_aliases import AppId_t, char, char_pointer, uint32, void_pointer
from ..constants import STEAMWORKS_BINARIES_PATH, STEAMWORKS_SDK_PATH
from ..exceptions import SteamworksError, UnsupportedPlatformError


class SteamworksApi:
    _api: CDLL

    def __init__(self, api_binary_path: str):
        """
        Loads the Steamworks API binary.
        :param api_binary_path: the path to the API binary
        """
        self._api = cdll.LoadLibrary(name=api_binary_path)

    def setup_types(self) -> None:
        """Sets up native function types."""
        # SteamAPI
        self._api.SteamAPI_Init.restype = ctypes.c_bool

        # SteamApps
        self._api.SteamAPI_SteamApps_v008.restype = void_pointer
        self._api.SteamAPI_ISteamApps_GetAppInstallDir.restype = uint32
        self._api.SteamAPI_ISteamApps_GetAppInstallDir.argtypes = (void_pointer, AppId_t, char_pointer, uint32)

        # SteamFriends
        self._api.SteamAPI_SteamFriends_v017.restype = void_pointer
        self._api.SteamAPI_ISteamFriends_GetPersonaName.restype = char_pointer
        self._api.SteamAPI_ISteamFriends_GetPersonaName.argtypes = (void_pointer,)

    def init(self) -> None:
        """
        Invoke SteamAPI_Init.
        :raises SteamworksError: if initialization failed
        """
        if not self._api.SteamAPI_Init():
            raise SteamworksError("SteamAPI initialization failed.")

    def _steam_apps(self) -> void_pointer:
        """
        Returns a pointer to the ISteamApps interface.
        :return: a pointer to ISteamApps
        :raises SteamworksError: if the API returned a null pointer
        """
        steam_apps: void_pointer = self._api.SteamAPI_SteamApps_v008()
        if not steam_apps:
            raise SteamworksError("The pointer to ISteamApps was null.")
        return steam_apps

    def _steam_friends(self) -> void_pointer:
        """
        Returns a pointer to the ISteamFriends interface.
        :return: a pointer to ISteamFriends
        :raises SteamworksError: if the API returned a null pointer
        """
        steam_friends: void_pointer = self._api.SteamAPI_SteamFriends_v017()
        if not steam_friends:
            raise SteamworksError("The pointer to ISteamFriends was null.")
        return steam_friends

    def get_install_directory(self, app_id: int) -> str:
        """
        Gets the installation path of the app id.
        :param app_id: the app id of the game
        :return: a path to the app directory
        :raises SteamworksError: if the installation path returned was empty
        :raises SteamworksError: if the API returned a null pointer (SteamworksApi._steam_apps)
        """
        # Get the ISteamApps pointer
        steam_apps: void_pointer = self._steam_apps()

        # Allocate a path buffer
        max_path_size: int = 260  # ctypes.wintypes.MAX_PATH
        install_directory_buffer: Array[char] = ctypes.create_string_buffer(max_path_size)

        # Invoke SteamAPI_ISteamApps_GetAppInstallDir
        bytes_written: uint32 = self._api.SteamAPI_ISteamApps_GetAppInstallDir(
            steam_apps,
            AppId_t(app_id),
            install_directory_buffer,
            len(install_directory_buffer))
        if not bytes_written:
            raise SteamworksError("ISteamApps::GetAppInstallDir returned zero bytes.")

        # Strip the trailing null bytes and return the installation path
        return install_directory_buffer.raw.rstrip(b"\x00").decode()

    def get_persona_name(self) -> str:
        """
        Gets the user's persona (display) name.
        :return: the user's display name
        """
        # Get the ISteamFriends pointer
        steam_friends: void_pointer = self._steam_friends()

        # Invoke SteamAPI_ISteamFriends_GetPersonaName
        persona_name: bytes = self._api.SteamAPI_ISteamFriends_GetPersonaName(steam_friends)
        return persona_name.decode()

    def shutdown(self) -> None:
        """Invoke SteamAPI_Shutdown."""
        self._api.SteamAPI_Shutdown()

    @staticmethod
    def prepare_and_initialize(api_binary_path: str, app_id: int) -> SteamworksApi:
        """
        This function will set up everything required to use the Steamworks API.
        :param api_binary_path: the path of the API binary
        :param app_id: the app id of the game
        :return: a new SteamworksApi instance
        :raises SteamworksError: if initialization failed (SteamworksApi.init)
        """
        # Setup the SteamAppId environment variable:
        #   the Steam API will attempt to resolve this
        #   environment variable because "steam_appid.txt"
        #   doesn't exist in the working directory.
        os.environ["SteamAppId"] = f"{app_id}"

        # Initialize the api
        api: SteamworksApi = SteamworksApi(api_binary_path=api_binary_path)
        api.setup_types()

        try:
            # Wrap SteamworksApi.init to guarantee that SteamAPI_Shutdown is invoked
            api.init()
        except SteamworksError:
            # Cleanup/shutdown the Steamworks API
            api.shutdown()
            raise

        # Return the api instance
        return api


def generate_api_binary_path() -> str:
    """
    Generates a path to the Steamworks API binary depending on the platform.
    :return: a path to the Steamworks API binary
    :raises UnsupportedPlatformError: if the current platform isn't supported
    """
    file_name: str
    bits: str = "64" if ctypes.sizeof(void_pointer) == 8 else ""
    match sys.platform:
        case "linux":
            file_name = f"libsteam_api{bits}.so"
        case "win32" | "cygwin":
            file_name = f"steam_api{bits}.dll"
        case "darwin":
            file_name = f"libsteam_api{bits}.dylib"
        case _:
            raise UnsupportedPlatformError(f"Unsupported platform: {sys.platform}")

    return os.path.join(STEAMWORKS_BINARIES_PATH, file_name)


def try_extract_steamworks_binaries() -> str:
    """
    Downloads the Steamworks SDK and extracts the API binary to the API path.
    :return: a path to the Steamworks API binary
    :raises SteamworksError: if the Steamworks SDK is missing from disk at STEAMWORKS_SDK_PATH
    :raises UnsupportedPlatformError: if the current platform isn't supported (generate_api_binary_path)
    """
    # Check if the binary has already been downloaded
    expected_api_binary_path: str = generate_api_binary_path()
    if os.path.exists(expected_api_binary_path):
        return expected_api_binary_path

    if not os.path.exists(STEAMWORKS_SDK_PATH):
        raise SteamworksError("Missing the Steamworks SDK.")

    # Extract the correct binary to the disk
    with ZipFile(file=STEAMWORKS_SDK_PATH) as zip_file:
        # Generate the binary file path
        file_name: str
        match (sys.platform, ctypes.sizeof(void_pointer) * 8):
            case ("linux", 32):
                file_name = "sdk/redistributable_bin/linux32/libsteam_api.so"
            case ("linux", 64):
                file_name = "sdk/redistributable_bin/linux64/libsteam_api.so"
            case ("win32" | "cygwin", 32):
                file_name = "sdk/redistributable_bin/steam_api.dll"
            case ("win32" | "cygwin", 64):
                file_name = "sdk/redistributable_bin/win64/steam_api64.dll"
            case ("darwin", 64):
                file_name = "sdk/redistributable_bin/osx/libsteam_api.dylib"
            case _:
                raise UnsupportedPlatformError(f"Unsupported platform: {sys.platform}")

        # Open the binary file and write it to disk
        with zip_file.open(name=file_name) as zip_binary_file:
            with open(expected_api_binary_path, mode="wb") as disk_binary_file:
                disk_binary_file.write(zip_binary_file.read())

    # Return the Steamworks API binary path
    return expected_api_binary_path


def fetch_hunt_attributes_path(steamworks_api: SteamworksApi, app_id: int) -> str:
    """
    Locates the game's install path and appends the attributes file path to it.
    :param steamworks_api: a Steamworks API instance
    :param app_id: the app id of the game
    :return: A path to the attributes file
    """
    install_directory: str = steamworks_api.get_install_directory(app_id=app_id)
    return install_directory + r"\user\profiles\default\attributes.xml"



import logging
import os.path
import statistics
import sys
import time
import xml.etree.ElementTree as ElementTree
from functools import partial

from colorama import Fore, Style, colorama_text

from hunt.attributes.parser import Match, Player, XmlElement, parse_match
from hunt.cli.arguments.parser import Config, parse_arguments
from hunt.cli.exit_codes import ExitCode
from hunt.constants import DATABASE_PATH, DATABASE_TEST_SERVER_PATH, HUNT_SHOWDOWN_APP_ID, \
    HUNT_SHOWDOWN_TEST_SERVER_APP_ID, MATCH_LOGS_PATH, RESOURCES_PATH, STEAMWORKS_BINARIES_PATH, STEAMWORKS_SDK_PATH
from hunt.database.client import Client as DatabaseClient
from hunt.exceptions import ParserError, SteamworksError, UnsupportedPlatformError
from hunt.filesystem.watchdog import FileWatchdog
from hunt.formats import format_mmr
from hunt.reward_constants import ASSISTS_CATEGORY
from hunt.steam.api import SteamworksApi, fetch_hunt_attributes_path, try_extract_steamworks_binaries


def setup_logger(config: Config) -> None:
    """
    Sets up the application logger.
    :param config: the configuration provided by the user
    """
    # Setup logging
    logging.basicConfig(format="[%(asctime)s, %(levelname)s] %(message)s",
                        datefmt="%H:%M", level=logging.INFO if not config.debug else logging.DEBUG,
                        stream=sys.stdout)

    def setup_level_color(level: int, color: str, padding: int = 8) -> None:
        """
        Adjusts a logging level name to add fancy color support.
        :param level: the logger level
        :param color: the color to use
        :param padding: the number of padded whitespace characters
        """
        logging.addLevelName(level, f"{color}{logging.getLevelName(level):>{padding}}{Style.RESET_ALL}")

    # Color config
    color_config: tuple[tuple[int, str], ...] = (
        (logging.DEBUG, Fore.CYAN),
        (logging.INFO, Fore.GREEN),
        (logging.WARNING, Fore.YELLOW),
        (logging.ERROR, Fore.RED),
        (logging.CRITICAL, Style.BRIGHT + Fore.RED))

    # Setup the color levels
    for logger_level, level_color in color_config:
        setup_level_color(logger_level, level_color)


def console_main() -> ExitCode:
    """
    The CLI entry point for the package.
    :return: an exit code.
    """
    # Initialize Colorama
    with colorama_text():
        # Parse the arguments and forward the app config to main
        config: Config = parse_arguments()

        # Setup the logger
        setup_logger(config)

        # Start the application.
        return main(config)


def main(config: Config) -> ExitCode:
    """
    Run the program and provide an exit code.
    :param config: the configuration provided by the user
    :return: an exit code.
    """
    try:
        os.makedirs(name=RESOURCES_PATH, exist_ok=True)  # Create the resource directory if it doesn't exist
        os.makedirs(name=MATCH_LOGS_PATH, exist_ok=True)  # Create the match logs directory if it doesn't exist
        os.makedirs(name=STEAMWORKS_BINARIES_PATH, exist_ok=True)  # Create the bin directory if it doesn't exist
    except OSError as exception:
        logging.critical("Failed to create create application-critical directories in the current working directory. "
                         "Are write permissions missing?")
        logging.debug(f"OS error: {exception=}")
        return ExitCode.FILESYSTEM_ERROR

    try:
        # Extract the Steamworks binaries to disk
        steamworks_api_path: str = try_extract_steamworks_binaries()
    except SteamworksError as exception:
        logging.critical("Failed to extract the Steamworks binaries, are you missing the Steamworks SDK?")
        logging.info(f"The Steamworks SDK should be located at: {STEAMWORKS_SDK_PATH!r}")
        logging.debug(f"Steamworks error: {exception=}")
        return ExitCode.STEAMWORKS_ERROR
    except UnsupportedPlatformError as exception:
        logging.critical("The current platform isn't supported.")
        logging.debug(f"Unsupported platform error: {exception=}")
        return ExitCode.UNSUPPORTED_PLATFORM

    try:
        # Initialize the Steamworks API
        app_id: int = HUNT_SHOWDOWN_APP_ID if not config.test_server else HUNT_SHOWDOWN_TEST_SERVER_APP_ID
        steamworks_api: SteamworksApi = SteamworksApi.prepare_and_initialize(steamworks_api_path, app_id=app_id)
    except SteamworksError as exception:
        logging.critical("A Steamworks API error occurred, is Steam running?")
        logging.debug(f"Steamworks error: {exception=}")
        return ExitCode.STEAMWORKS_ERROR
    else:
        logging.info("Steamworks API initialized.")

    # Locate the attributes file
    attributes_path: str = fetch_hunt_attributes_path(steamworks_api, app_id=app_id)
    assert os.path.exists(attributes_path), "Attributes file does not exist."

    database: DatabaseClient
    database_path: str = DATABASE_PATH if not config.test_server else DATABASE_TEST_SERVER_PATH
    with DatabaseClient(file_path=database_path) as database:
        # Set up a file watcher to listen for changes on the attributes file
        file_watchdog: FileWatchdog = FileWatchdog(
            file_path=attributes_path,
            callback=partial(attributes_file_modified,
                             database=database, steamworks_api=steamworks_api, config=config))
        file_watchdog.start()

        # Inform the user that the program has started
        logging.info("Watching for matches, good luck and have fun!")
        try:
            while True:
                time.sleep(1)
        except KeyboardInterrupt:
            # Stop the file watcher if a keyboard interrupt is received
            file_watchdog.stop()
        file_watchdog.join()

    # Cleanup/shutdown the Steamworks API
    steamworks_api.shutdown()

    # Signal to the user that we're shutting down
    logging.info("Shutting down.")
    return ExitCode.SUCCESS


def attributes_file_modified(file_path: str,
                             database: DatabaseClient, steamworks_api: SteamworksApi, config: Config) -> None:
    """
    Invoked when the attributes file is modified;
      Parses the match data from the attributes file and
      saves it to disk.
    :param file_path: the path of the file to parse
    :param database: a DatabaseClient instance
    :param steamworks_api: a SteamworksApi instance
    :param config: the configuration provided by the user
    """
    # Read the file contents
    with open(file_path, "rb") as file:
        file_contents: bytes = file.read()

    # If the file contents are empty, skip parsing
    if not file_contents:
        return

    try:
        # Attempt to parse the attributes file
        parsed_attributes: XmlElement = ElementTree.fromstring(file_contents)
    except ElementTree.ParseError as exception:
        # Skip the update
        logging.error("Failed to parse the attributes file.")
        logging.debug(f"Failed to parse the attributes file: {exception=}")
        return

    # Parse the teams from the attributes file
    try:
        match: Match = parse_match(root=parsed_attributes, steam_name=steamworks_api.get_persona_name())
    except SteamworksError as exception:
        logging.debug(f"Failed to get the user's display name: {exception=}")
        return
    except ParserError as exception:
        logging.debug(f"Failed to parse the attributes file: {exception=}")
        return

    # Save match data to disk
    if match.try_save_to_file(database=database):
        return  # Skip printing an already existing entry

    # Print useful data from the match
    log_match_data(match, log_statistical_data=config.statistics)


def log_match_data(match: Match, log_statistical_data: bool) -> None:
    """
    Logs interesting data about the match such as:
      - rewards collected from the match,
      - players in the team, and
      - enemy players that the player interacted with.
    :param match: a parsed Match instance
    :param log_statistical_data: True if statistical match data should be presented to the user
    """
    # Log statistical data
    def _log_stats() -> None:
        logging.info("Statistics:")

        # Player statistics
        players: tuple[Player, ...] = tuple(player for team in match.teams for player in team.players)
        kills: int = sum(player.downed_by_me + player.killed_by_me for player in players if not player.is_partner)
        deaths: int = sum(player.downed_me + player.killed_me for player in players if not player.is_partner)
        assists: int = sum(entry.amount for entry in match.entries if entry.category == ASSISTS_CATEGORY)

        # Avoid ZeroDivisionError
        if not deaths:
            deaths += 1

        logging.info(f"  Bloodline rank: {match.bloodline_rank}")
        logging.info(f"  Amount of players: {len(players)}")

        kd_ratio: float = kills / deaths
        kda_ratio: float = (kills + assists) / deaths
        if kd_ratio or kda_ratio:
            logging.info(f"  KD: {kd_ratio:.2f}, KDA: {kda_ratio:.2f}")

        # Match statistics
        mmr_data_set: tuple[int, ...] = tuple(player.mmr for player in players)
        logging.info(f"  MMR: average: {format_mmr(int(statistics.mean(mmr_data_set)))} ± "
                     f"{int(statistics.pstdev(mmr_data_set))}, "
                     f"lowest: {format_mmr(min(mmr_data_set))}, "
                     f"median: {format_mmr(int(statistics.median_high(mmr_data_set)))}, "
                     f"highest: {format_mmr(max(mmr_data_set))}")
    if log_statistical_data:
        _log_stats()

    # Log interesting rewards
    def _log_rewards() -> None:
        logging.info("Rewards:")

        # Hunt dollars
        total_hunt_dollars: int = match.rewards.bounty + match.rewards.hunt_dollars
        if total_hunt_dollars:
            logging.info(f"  Received {total_hunt_dollars} hunt dollar(s).")

        # Bloodbonds
        if match.rewards.bloodbonds:
            logging.info(f"  Received {match.rewards.bloodbonds} bloodbond(s).")

        # Hunter-related rewards
        total_xp: int = match.rewards.xp + match.rewards.hunter_xp + match.rewards.bounty * 4
        if total_xp:
            logging.info(f"  Received {total_xp} hunter XP.")
        if match.rewards.hunter_levels:
            logging.info(f"  Received {match.rewards.hunter_levels} perk points.")
        if match.rewards.upgrade_points:
            logging.info(f"  Collected {match.rewards.upgrade_points} upgrade points.")

        # Bloodline XP
        if match.rewards.bloodline_xp:
            logging.info(f"  Collected {match.rewards.bloodline_xp} bloodline XP.")

        # Event points
        if match.rewards.event_points:
            logging.info(f"  Collected {match.rewards.event_points} event points.")
    # Skip logging if there were no rewards
    if match.rewards:
        _log_rewards()

    # Log players
    def _log_players() -> None:
        # Log information about the local team
        logging.info("Team:")
        for player in (player for team in match.teams for player in team.players if team.own_team):
            logging.info(f"  {player.format_name(is_local_player=player.name == match.player_name)}")

        # Log information about the players the local player interacted with
        players: tuple[Player, ...] = tuple(player for team in match.teams for player in team.players)
        if any(player.downed_by_me or player.downed_me or player.killed_by_me or player.killed_me
               for player in players if not player.is_partner):
            logging.info("Enemies:")
            for player in players:
                if player.downed_by_me or player.killed_by_me:
                    logging.info(f"  {player.format_kills()}")
                if player.downed_me or player.killed_me:
                    logging.info(f"  {player.format_deaths()}")
    _log_players()


if __name__ == "__main__":
    raise SystemExit(console_main())



import ctypes
from typing import TypeAlias


# Native types
char: TypeAlias = ctypes.c_char
char_pointer: TypeAlias = ctypes.c_char_p
void_pointer: TypeAlias = ctypes.c_void_p
uint32: TypeAlias = ctypes.c_uint32

# Steamworks API types
AppId_t: TypeAlias = uint32



from .match import Accolade, Entry, Match, Rewards, Team
from .team import Player, SerializableTeam
from .xml.elements import XmlElement, get_element_value
from ..reward_constants import BLOODBONDS_CATEGORY, BLOODLINE_DESCRIPTOR_NAME, BOUNTY_CATEGORIES, \
    HUNTER_LEVELS_CATEGORY, HUNTER_XP_DESCRIPTOR_NAME, HUNTER_XP_REWARD_TYPE, HUNT_DOLLARS_CATEGORY, \
    UPGRADE_POINTS_DESCRIPTOR_NAME, XP_CATEGORIES


def _calculate_rewards(accolades: tuple[Accolade, ...], entries: tuple[Entry, ...],
                       hunt_dollar_bonus: int, hunter_xp_bonus: int) -> Rewards:
    """
    Calculates all the rewards collected from a match.
    :param accolades: a tuple of Accolade instances
    :param entries: a tuple of Entry instances
    :return: a Rewards instance
    """
    generated_bloodbonds: int = sum(accolade.generated_bloodbonds for accolade in accolades)
    bounty: int = sum(entry.reward_size for entry in entries if entry.category in BOUNTY_CATEGORIES)
    xp: int = sum(entry.reward_size for entry in entries if entry.category in XP_CATEGORIES)
    hunt_dollars: int = sum(entry.reward_size for entry in entries if entry.category == HUNT_DOLLARS_CATEGORY)
    bloodbonds: int = sum(entry.reward_size for entry in entries if entry.category == BLOODBONDS_CATEGORY)
    hunter_xp: int = sum(entry.reward_size for entry in entries
                         if entry.descriptor_name == HUNTER_XP_DESCRIPTOR_NAME
                         and entry.reward_type == HUNTER_XP_REWARD_TYPE)
    hunter_levels: int = sum(entry.reward_size for entry in entries if entry.category == HUNTER_LEVELS_CATEGORY)
    upgrade_points: int = sum(entry.reward_size for entry in entries
                              if entry.descriptor_name == UPGRADE_POINTS_DESCRIPTOR_NAME)
    bloodline_xp: int = sum(entry.reward_size for entry in entries
                            if entry.descriptor_name == BLOODLINE_DESCRIPTOR_NAME)
    event_points: int = sum(accolade.event_points for accolade in accolades)

    return Rewards(bounty, xp + hunter_xp_bonus, hunt_dollars + hunt_dollar_bonus, generated_bloodbonds + bloodbonds,
                   hunter_xp, hunter_levels, upgrade_points, bloodline_xp, event_points)


def parse_match(root: XmlElement, steam_name: str) -> Match:
    """
    Parse the element tree for match data.
    :param root: the root element tree
    :param steam_name: the user's display name
    :return: a Match object
    :raises ParserError: from get_element_value
    """
    accolades: list[Accolade] = []
    entries: list[Entry] = []

    # Determine the expected number of accolades and entries to iterate
    accolades_count: int = get_element_value(root, "MissionBagNumAccolades", result_type=int)
    entries_count: int = get_element_value(root, "MissionBagNumEntries", result_type=int)

    # Parse and store the accolades
    for i in range(accolades_count):
        accolades.append(Accolade.deserialize(root, accolade_id=i))

    # Parse and store the entries
    for i in range(entries_count):
        entries.append(Entry.deserialize(root, entry_id=i))

    bloodline_rank: int = get_element_value(root, "Unlocks/UnlockRank", result_type=int)
    hunt_dollar_bonus: int = get_element_value(root, "MissionBagFbeGoldBonus", result_type=int)
    hunter_xp_bonus: int = get_element_value(root, "MissionBagFbeHunterXpBonus", result_type=int)
    is_hunter_dead: bool = get_element_value(root, "MissionBagIsHunterDead", result_type=bool)
    is_quickplay: bool = get_element_value(root, "MissionBagIsQuickPlay", result_type=bool)
    region: str = get_element_value(root, "Region")
    secondary_region: str = get_element_value(root, "SecondaryRegion")

    accolades_tuple: tuple[Accolade, ...] = tuple(accolades)
    entries_tuple: tuple[Entry, ...] = tuple(entries)
    return Match(steam_name, bloodline_rank, is_hunter_dead, is_quickplay, region, secondary_region,
                 accolades_tuple, entries_tuple,
                 _calculate_rewards(accolades_tuple, entries_tuple, hunt_dollar_bonus, hunter_xp_bonus),
                 parse_teams(root=root))


def parse_teams(root: XmlElement) -> tuple[Team, ...]:
    """
    Parse the element tree for the available teams.
    :param root: the root element tree
    :return: a tuple of Team objects
    :raises ParserError: from get_element_value
    """
    teams: list[Team] = []

    # Determine the expected team count and iterate over the teams
    expected_team_count: int = get_element_value(root, "MissionBagNumTeams", result_type=int)
    for i in range(expected_team_count):
        # Parse each team
        parsed_team: SerializableTeam = SerializableTeam.deserialize(root, team_id=i)

        players: list[Player] = []
        # Parse each player from the team
        for j in range(parsed_team.players_count):
            players.append(Player.deserialize(root, team_id=i, player_id=j))

        teams.append(Team(parsed_team.handicap, parsed_team.is_invite, parsed_team.mmr, parsed_team.own_team,
                          tuple(players)))
    return tuple(teams)



import os.path
from typing import Callable

from watchdog.events import FileModifiedEvent, FileSystemEvent, FileSystemEventHandler
from watchdog.observers import Observer


class FileWatchdog(FileSystemEventHandler):
    file_path: str
    callback: Callable[..., None]
    _observer: Observer

    def __init__(self, file_path: str, callback: Callable[..., None]):
        """
        Initialize the class.
        :param file_path: the file path to monitor for changes
        :param callback: the callback to invoke when changes are detected
        """
        self.file_path = os.path.realpath(file_path)
        self.callback = callback

        # https://github.com/python/mypy/issues/10757
        self._observer = Observer()  # type: ignore[no-untyped-call]
        self._observer.schedule(event_handler=self, path=os.path.dirname(file_path))  # type: ignore[no-untyped-call]

    def start(self) -> None:
        """Start the observer."""
        # https://github.com/python/mypy/issues/10757
        self._observer.start()  # type: ignore[no-untyped-call]

    def join(self) -> None:
        """Wait until the observer thread terminates."""
        self._observer.join()

    def stop(self) -> None:
        """Stop the observer."""
        # https://github.com/python/mypy/issues/10757
        self._observer.stop()  # type: ignore[no-untyped-call]

    def on_modified(self, event: FileSystemEvent) -> None:
        """
        Invoked when a modified event is generated.
        :param event: a watchdog event
        """
        if not isinstance(event, FileModifiedEvent) or os.path.realpath(event.src_path) != self.file_path:
            return
        self.callback(self.file_path)



from __future__ import annotations

from dataclasses import dataclass

from .player import Player
from .xml.serializable import MappingGenerator, Serializable, XmlElement


@dataclass(frozen=True)
class SerializableTeam(Serializable):
    handicap: int
    is_invite: bool
    mmr: int
    players_count: int
    own_team: bool

    # Serialization
    @staticmethod
    def _generate_prefix(team_id: int) -> str:
        """
        Generate a prefix for the class.
        :param team_id: the id of the team
        :return: a MissionBagEntry prefix
        """
        return f"MissionBagTeam_{team_id}"

    @staticmethod
    def _data_mappings() -> MappingGenerator:
        """
        Yield each variable name and its corresponding attribute suffix.
        :return: a generator which yields the name, value
        """
        # Yield the values
        yield "handicap", "handicap"
        yield "is_invite", "isinvite"
        yield "mmr", "mmr"
        yield "players_count", "numplayers"
        yield "own_team", "ownteam"

    # noinspection PyMethodOverriding
    def serialize(self, root: XmlElement, *, team_id: int) -> None:  # type: ignore[override]
        """
        Serialize a SerializableTeam instance.
        :param root: the root element to serialize into
        :param team_id: the id of the team
        """
        # Generate the prefix
        prefix: str = SerializableTeam._generate_prefix(team_id)

        # Serialize the data
        super().serialize(root, prefix)

    # noinspection PyMethodOverriding
    @classmethod
    def deserialize(cls, root: XmlElement, team_id: int) -> SerializableTeam:  # type: ignore[override]
        """
        Deserialize a series of elements into a SerializableTeam instance.
        :param root: the root element to serialize into
        :param team_id: the id of the team
        :return: a serialized SerializableTeam instance
        """
        # Generate the prefix
        prefix: str = SerializableTeam._generate_prefix(team_id)

        # Deserialize the data
        return super(cls, cls).deserialize(root, prefix)


@dataclass(frozen=True)
class Team:
    handicap: int
    is_invite: bool
    mmr: int
    own_team: bool
    players: tuple[Player, ...]

    # Serialization support
    def to_serializable_team(self) -> SerializableTeam:
        """
        Construct a SerializableTeam from the current instance.
        :return: a populated SerializableTeam instance
        """
        return SerializableTeam(self.handicap, self.is_invite, self.mmr, len(self.players), self.own_team)



from argparse import ArgumentParser, Namespace

from ..config import Config


def setup_argument_parser() -> ArgumentParser:
    """
    Sets up an argument parser with all supported commands.
    :return: an ArgumentParser instance
    """
    argument_parser: ArgumentParser = ArgumentParser(add_help=False, exit_on_error=False)

    # Debug variables
    argument_parser.add_argument("--debug", action="store_true")

    # Test server support
    argument_parser.add_argument("--test-server", action="store_true")

    # Statistics
    argument_parser.add_argument("--statistics", action="store_true")

    return argument_parser


def parse_arguments() -> Config:
    """
    Parses command line arguments into a Config instance.
    :return: a Config instance
    """
    # Parse any provided arguments
    argument_parser: ArgumentParser = setup_argument_parser()
    arguments: Namespace = argument_parser.parse_args()

    # Return a Config instance
    return Config(arguments.debug, arguments.test_server, arguments.statistics)



from __future__ import annotations

from contextlib import closing
from dataclasses import dataclass
from sqlite3 import Connection, Cursor, connect as sqlite3_connect
from types import TracebackType

from ..constants import DATABASE_TABLE_QUERIES


@dataclass(kw_only=True)
class Client:
    file_path: str
    _connection: Connection | None = None

    def __post_init__(self) -> None:
        """Setup the database connection."""
        self._connection = sqlite3_connect(f"file:{self.file_path}", check_same_thread=False, uri=True)
        self._setup_database()

    def _setup_database(self) -> None:
        """Sets up the database by creating the required tables."""
        cursor: Cursor
        with closing(self.cursor()) as cursor:
            # Setup each table
            for table_query in DATABASE_TABLE_QUERIES:
                cursor.execute(table_query)
        self.save()

    def cursor(self) -> Cursor:
        """Returns a new Cursor instance."""
        assert self._connection is not None
        return self._connection.cursor()

    def save(self) -> None:
        """Commit the changes to disk."""
        assert self._connection is not None
        self._connection.commit()

    def close(self) -> None:
        """Closes the connection."""
        assert self._connection is not None
        self._connection.close()

    # Context manager support
    def __enter__(self) -> Client:
        """Return self when entering the scope."""
        return self

    def __exit__(self, exc_type: type[BaseException] | None, exc_val: BaseException | None,
                 exc_tb: TracebackType | None) -> None:
        """Close the database connection when exiting the scope."""
        self.close()



from __future__ import annotations

from dataclasses import dataclass

from colorama import Fore, Style

from .xml.serializable import MappingGenerator, Serializable, XmlElement
from ..formats import format_mmr


@dataclass(frozen=True)
class Player(Serializable):
    name: str
    bounties_extracted: int
    bounties_picked_up: int
    downed_by_me: int
    downed_by_teammate: int
    downed_me: int
    downed_teammate: int
    had_wellspring: bool
    is_partner: bool
    is_soul_survivor: bool
    killed_by_me: int
    killed_by_teammate: int
    killed_me: int
    killed_teammate: int
    mmr: int
    profile_id: int
    proximity_to_me: bool
    proximity_to_teammate: bool
    skillbased: bool
    team_extraction: bool

    # Formatting
    def format_name(self, color_prefix: str = Fore.GREEN, is_local_player: bool = False) -> str:
        """
        Generate a formatted name.
        :param color_prefix: the color to prefix the name with
        :param is_local_player: the Player instance is the instance of the local player
        :return: a color formatted string
        """
        local_player_marker: str = " (you)" if is_local_player else ""
        return f"{color_prefix}{self.name}{Style.RESET_ALL} ({format_mmr(self.mmr)}){local_player_marker}"

    def format_kills(self) -> str:
        """
        Generate a formatted name with the amount of times this player was killed by the local player.
        :return: a color formatted string
        """
        kills: int = self.downed_by_me + self.killed_by_me
        kill_count: str = f" {kills}x" if kills > 1 else ""
        return f"Killed {self.format_name()}{kill_count}"

    def format_deaths(self) -> str:
        """
        Generate a formatted name with the amount of times this player killed the local player.
        :return: a color formatted string
        """
        deaths: int = self.downed_me + self.killed_me
        death_count: str = f" {deaths}x" if deaths > 1 else ""
        return f"Died to {self.format_name(color_prefix=Fore.RED)}{death_count}"

    # Serialization
    @staticmethod
    def _generate_prefix(team_id: int, player_id: int) -> str:
        """
        Generate a prefix for the class.
        :param team_id: the team id of the player's team
        :param player_id: the id of the player
        :return: a MissionBagPlayer prefix
        """
        return f"MissionBagPlayer_{team_id}_{player_id}"

    # noinspection PyMethodOverriding
    @staticmethod
    def _data_mappings() -> MappingGenerator:
        """
        Yield each variable name and its corresponding attribute suffix.
        :return: a generator which yields the name, value
        """
        # Yield the values
        yield "name", "blood_line_name"
        yield "bounties_extracted", "bountyextracted"
        yield "bounties_picked_up", "bountypickedup"
        yield "downed_by_me", "downedbyme"
        yield "downed_by_teammate", "downedbyteammate"
        yield "downed_me", "downedme"
        yield "downed_teammate", "downedteammate"
        yield "had_wellspring", "hadWellspring"
        yield "is_partner", "ispartner"
        yield "is_soul_survivor", "issoulsurvivor"
        yield "killed_by_me", "killedbyme"
        yield "killed_by_teammate", "killedbyteammate"
        yield "killed_me", "killedme"
        yield "killed_teammate", "killedteammate"
        yield "mmr", "mmr"
        yield "profile_id", "profileid"
        yield "proximity_to_me", "proximitytome"
        yield "proximity_to_teammate", "proximitytoteammate"
        yield "skillbased", "skillbased"
        yield "team_extraction", "teamextraction"

    # noinspection PyMethodOverriding
    def serialize(self, root: XmlElement, *, team_id: int, player_id: int) -> None:  # type: ignore[override]
        """
        Serialize a Player instance.
        :param root: the root element to serialize into
        :param team_id: the team id of the player
        :param player_id: the id of the player
        """
        # Generate the prefix
        prefix: str = Player._generate_prefix(team_id, player_id)

        # Serialize the data
        super().serialize(root, prefix)

    # noinspection PyMethodOverriding
    @classmethod
    def deserialize(cls, root: XmlElement, *, team_id: int, player_id: int) -> Player:  # type: ignore[override]
        """
        Deserialize a series of elements into a Player instance.
        :return: a serialized Player instance
        """
        # Generate the prefix
        prefix: str = Player._generate_prefix(team_id, player_id)

        # Deserialize the data
        return super(cls, cls).deserialize(root, prefix)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['hunt',
 'hunt.attributes',
 'hunt.attributes.xml',
 'hunt.cli',
 'hunt.cli.arguments',
 'hunt.database',
 'hunt.filesystem',
 'hunt.steam']

package_data = \
{'': ['*']}

install_requires = \
['colorama>=0.4.5,<0.5.0', 'watchdog>=2.1.9,<3.0.0']

entry_points = \
{'console_scripts': ['hunt-match-telemetry-cli = hunt.cli.app:console_main']}

setup_kwargs = {
    'name': 'hunt-match-telemetry',
    'version': '1.7.2',
    'description': 'Automatically extract match data from Hunt: Showdown matches.',
    'long_description': '[![Run Tests CI](https://github.com/anthonyprintup/hunt-match-telemetry/actions/workflows/run-tests.yaml/badge.svg)](https://github.com/anthonyprintup/hunt-match-telemetry/actions/workflows/run-tests.yaml)\n\n# Hunt Match Telemetry\nThis tool is intended to be used for automatically logging match data provided by the game.\n\n# Installation\n- Create a new Python 3.11 [virtual environment](https://docs.python.org/3/library/venv.html) and activate it, <sup><sub>(optional)<sub/></sup>\n- run `pip install hunt-match-telemetry` to install the package,\n- download the [Steamworks SDK](https://partner.steamgames.com/downloads/steamworks_sdk.zip) and place it in `./resources/steam` as `steamworks_sdk.zip`.\n\n# Instructions\n- Run the CLI version of the package by executing `hunt-match-telemetry-cli` in your preferred terminal,\n- join a match from the game,\n- finish the game (extract, die, etc.),\n- return to the lobby screen (or any UI element that updates the last match information).\n\n# Screenshots\n<!--suppress CheckImageSize, HtmlDeprecatedAttribute -->\n<p align="center">\n    Console log:\n    <br/>\n    <img alt="Console Log" src="https://github.com/anthonyprintup/hunt-match-telemetry/blob/main/assets/console_log_example.png?raw=true" />\n</p>\n<p align="center">\n    Match log:\n    <br/>\n    <img alt="Match Log" src="https://github.com/anthonyprintup/hunt-match-telemetry/blob/main/assets/match_log_example.png?raw=true" />\n</p>\n<p align="center">\n    Player log:\n    <br/>\n    <img alt="Player Log" src="https://github.com/anthonyprintup/hunt-match-telemetry/blob/main/assets/player_log_example.png?raw=true" width="715"/>\n</p>\n\n# Notice of Non-Affiliation and Disclaimer\nWe are not affiliated, associated, authorized, endorsed by, or in any way officially connected with Crytek GmbH, or any of its subsidiaries or its affiliates.\n\nThe official Crytek GmbH website can be found at https://www.crytek.com/.\nThe official Hunt: Showdown website can be found at https://www.huntshowdown.com/.\n\nAll product and company names are the registered trademarks of their original owners. The use of any trade name or trademark is for identification and reference purposes only and does not imply any association with the trademark holder of their product brand.\n',
    'author': 'Anthony Printup',
    'author_email': 'anthony@printup.io',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/anthonyprintup/hunt-match-telemetry',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



import os.path

HUNT_SHOWDOWN_APP_ID: int = 594650  # Hunt: Showdown App ID
HUNT_SHOWDOWN_TEST_SERVER_APP_ID: int = 770720  # Hunt: Showdown (Test Server) App ID

# The path (directory) to where match logs are stored
WORKING_DIRECTORY: str = os.getcwd()
RESOURCES_PATH: str = os.path.realpath(os.path.join(WORKING_DIRECTORY, "resources"))
MATCH_LOGS_PATH: str = os.path.join(RESOURCES_PATH, "logs")

# Steam
STEAMWORKS_BINARIES_PATH: str = os.path.join(RESOURCES_PATH, "steam")
STEAMWORKS_SDK_PATH: str = os.path.join(STEAMWORKS_BINARIES_PATH, "steamworks_sdk.zip")

# Formatting
STAR_SYMBOL: str = "★"
MMR_RANGES: tuple[int, ...] = (0, 2000, 2300, 2600, 2750, 3000)

# Database
DATABASE_PATH: str = os.path.join(RESOURCES_PATH, "match_data.db")
DATABASE_TEST_SERVER_PATH: str = os.path.join(RESOURCES_PATH, "match_data_ts.db")


# Helper function to generate create table queries
def _create_table_helper(table_name: str, fields: tuple[str, ...]) -> str:
    return f"CREATE TABLE IF NOT EXISTS {table_name} ({', '.join(fields)})"


# Tables
_PLAYER_LOG_COLUMNS: tuple[str, ...] = ("id INTEGER PRIMARY KEY",
                                        "profile_id INTEGER UNIQUE",
                                        "name TEXT NOT NULL", "mmr INTEGER DEFAULT 0 NOT NULL",
                                        "kills INTEGER DEFAULT 0 NOT NULL", "deaths INTEGER DEFAULT 0 NOT NULL",
                                        "encounters INTEGER DEFAULT 0 NOT NULL")
DATABASE_TABLE_QUERIES: tuple[str, ...] = (
    _create_table_helper("data_hashes", ("id INTEGER PRIMARY KEY", "hash varchar(64) UNIQUE", "path TEXT NOT NULL")),
    _create_table_helper("player_log_bountyhunt", _PLAYER_LOG_COLUMNS),
    _create_table_helper("player_log_quickplay", _PLAYER_LOG_COLUMNS))



from __future__ import annotations

from dataclasses import dataclass

from .xml.serializable import MappingGenerator, Serializable, XmlElement


@dataclass(frozen=True)
class Entry(Serializable):
    amount: int
    category: str
    descriptor_name: str
    descriptor_score: int
    descriptor_type: int
    reward_type: int
    reward_size: int

    # Serialization
    @staticmethod
    def _generate_prefix(entry_id: int) -> str:
        """
        Generate a prefix for the class.
        :param entry_id: the id of the entry
        :return: a MissionBagEntry prefix
        """
        return f"MissionBagEntry_{entry_id}"

    @staticmethod
    def _data_mappings() -> MappingGenerator:
        """
        Yield each variable name and its corresponding attribute suffix.
        :return: a generator which yields the name, value
        """
        # Yield the values
        yield "amount", "amount"
        yield "category", "category"
        yield "descriptor_name", "descriptorName"
        yield "descriptor_score", "descriptorScore"
        yield "descriptor_type", "descriptorType"
        yield "reward_type", "reward"
        yield "reward_size", "rewardSize"

    # noinspection PyMethodOverriding
    def serialize(self, root: XmlElement, *, entry_id: int) -> None:  # type: ignore[override]
        """
        Serialize an Entry instance.
        :param root: the root element to serialize into
        :param entry_id: the id of the entry
        """
        # Generate the prefix
        prefix: str = Entry._generate_prefix(entry_id)

        # Serialize the data
        super().serialize(root, prefix)

    # noinspection PyMethodOverriding
    @classmethod
    def deserialize(cls, root: XmlElement, entry_id: int) -> Entry:  # type: ignore[override]
        """
        Deserialize a series of elements into an Entry instance.
        :param root: the root element to serialize into
        :param entry_id: the id of the entry
        :return: a serialized Entry instance
        """
        # Generate the prefix
        prefix: str = Entry._generate_prefix(entry_id)

        # Deserialize the data
        return super(cls, cls).deserialize(root, prefix)



import json
import os.path
from dataclasses import asdict, dataclass
from datetime import datetime
from hashlib import sha256

from .accolade import Accolade
from .entry import Entry
from .rewards import Rewards
from .team import Team
from ..constants import MATCH_LOGS_PATH
from ..database.queries import DatabaseClient, data_hash_exists, insert_match_hash, update_player_data


@dataclass(frozen=True)
class Match:
    player_name: str
    bloodline_rank: int
    is_hunter_dead: bool
    is_quickplay: bool
    region: str
    secondary_region: str
    accolades: tuple[Accolade, ...]
    entries: tuple[Entry, ...]
    rewards: Rewards
    teams: tuple[Team, ...]

    def generate_file_path(self, time: datetime | None = None) -> str:
        """
        Generates a file path for the match.
        :param time: the time to use in this context
        :return: the file path for the match data
        """
        if time is None:
            time = datetime.now()
        return os.path.join(MATCH_LOGS_PATH, f"{time.year}-{time.month:02d}-{time.day:02d}",
                                             "quickplay" if self.is_quickplay else "bounty_hunt",
                                             f"{time.hour:02d}-{time.minute:02d}-{time.second:02d}.json")

    def generate_hash(self) -> str:
        """
        Generates a sha256 hash of the match.
        Some variables are removed to prevent duplicates.
        :return: a sha256 hash digest
        """
        # Convert the current instance to a dictionary
        match_data: dict = asdict(self)

        # Remove certain variables to prevent entry spamming
        del match_data["bloodline_rank"]
        del match_data["region"]
        del match_data["secondary_region"]

        return sha256(json.dumps(match_data).encode()).hexdigest()

    def try_save_to_file(self, database: DatabaseClient) -> bool:
        """
        Converts the match data to json and saves it to the file path,
          if the match data hasn't already been saved.
        :param database: a DatabaseClient instance
        :return: True if this entry already exists in the database, otherwise False.
        """
        # Generate a datetime instance
        current_time: datetime = datetime.now()

        # Generate the match hash
        match_hash: str = self.generate_hash()

        # Check if the hash already exists in the database to prevent duplicates
        if data_hash_exists(database, match_hash=match_hash):
            return True

        # Generate the file path
        generated_file_path: str = self.generate_file_path(time=current_time)

        # Save the hash to the database
        insert_match_hash(database, match_hash=match_hash, file_path=generated_file_path)

        # Update the player log
        for player in (player for team in self.teams for player in team.players):
            update_player_data(database, profile_id=player.profile_id, name=player.name, mmr=player.mmr,
                               kills=player.killed_by_me + player.downed_by_me,
                               deaths=player.killed_me + player.downed_me,
                               is_quickplay=self.is_quickplay)

        # Create the directories
        directory_path: str = os.path.dirname(generated_file_path)
        os.makedirs(name=directory_path, exist_ok=True)

        # Generate the match data as JSON
        match_data: str = json.dumps(self, indent=2, default=vars)

        # Save the data to a file
        with open(generated_file_path, mode="w") as file:
            file.write(match_data)
        return False
